[project]
name = "{{ project_name }}"
requires-python = ">=3.9,<3.13"
version = "0.1.0"
{{ dependencies }}

[project.entry-points]
"dagster.components" = { {{ project_name }} = "{{ project_name }}.lib"}

[dependency-groups]
{{ dev_dependencies }}

[build-system]
requires = ["setuptools"]
build-backend = "setuptools.build_meta"

[tool.dagster]
module_name = "{{ project_name }}.definitions"
code_location_name = "{{ code_location_name }}"

[tool.dg]
is_code_location = true
is_component_lib = true

[tool.setuptools.packages.find]
exclude=["{{ project_name }}_tests"]

{{ uv_sources }}
